import axinite as ax
import numpy as np

class Body(ax.Body):
    """
    A class that represents a body in the simulation.

    Attributes:
        name (str): The body's name.
        mass (np.float64): The mass of the body in kilograms.
        position (np.ndarray): The initial position of the body (in vector form).
        velocity (np.ndarray): The initial velocity of the body (in vector form).
        radius (np.float64): The radius of the body in meters.
        color (str): The color of the body.
        light (bool): Whether the body should give off light.
        retain (int): How many points the body should retain on its trail.
        radius_multiplier (int): A multiplier to be applied to the radius.
    """

    def __init__(self, name: str, mass: np.float64, limit: np.float64, delta: np.float64, position: np.ndarray = None, velocity: np.ndarray = None, radius: np.float64 = -1, color: str = "", light: bool = False, retain: int = None, radius_multiplier: int = 1):
        """
        Initializes a new Body object.

        Args:
            name (str): The body's name.
            mass (np.float64): The mass of the body in kilograms.
            limit (np.float64): The length of the simulation in seconds.
            delta (np.float64): The frequency at which the simulation should be computed in seconds.
            position (np.ndarray, optional): The initial position of the body (in vector form). Defaults to None.
            velocity (np.ndarray, optional): The initial velocity of the body (in vector form). Defaults to None.
        """
        if position is not None:
            position = np.array(position, dtype=np.float64)
        if velocity is not None:
            velocity = np.array(velocity, dtype=np.float64)
        mass = np.float64(mass)
        limit = np.float64(limit)
        delta = np.float64(delta)

        self.mass: np.float64 = mass
        "The mass of the body in kilograms."

        self.name: str = name
        "The name of the body."

        self._inner = ax._body(limit, delta, name, mass)

        if position is not None: self._inner["r"][0] = position
        if velocity is not None: self._inner["v"][0] = velocity

        self._inner["n"] = name
        self._inner["m"] = mass

        self.radius: np.float64 = radius
        "The radius of the body in meters."

        self.color: str = color
        "The color of the body."

        self.light: bool = light
        "Whether the body should give off light."

        self.retain: int = retain
        "How many points the body should retain on it's trail."

        self.radius_multiplier: int = radius_multiplier
        "A multiplier which was applied to the radius."

    def r(self, t: np.float64) -> np.ndarray:
        """Returns the position of the body at a specific time.

        Args:
            t (np.float64): The time to get the position at.

        Returns:
            np.ndarray: The position of the body at the time.
        """
        return self._inner["r"][int(t)]
    
    def v(self, t: np.float64) -> np.ndarray:
        """Returns the velocity of the body at a specific time.

        Args:
            t (np.float64): The time to get the velocity at.

        Returns:
            np.ndarray: The velocity of the body at the time.
        """
        return self._inner["v"][int(t)]


from vpython import *
import axinite.tools as axtools
from itertools import cycle
import signal

def show(_args: axtools.AxiniteArgs, frontend: 'function') -> None:
    """Statically display the bodies in the simulation.

    Args:
        _args (axtools.AxiniteArgs): The simulation parameters.
        frontend (function): The frontend function.
    """

    args = _args
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
        args.retain = 200

    for body in args.bodies:
        frontend[0](body)

    signal.signal(signal.SIGINT, lambda *args, **kwargs: frontend[2])
    
    if frontend[1] != None: 
        frontend[1]()


import axinite as ax
import axinite.analysis as axana
import numpy as np
from numba import jit

def rocket_autopilot(destination: np.ndarray, body: ax.Body, _bodies: np.ndarray, speed_max: np.float64, 
                     force_max: np.float64, turn_rate: np.float64, time: int) -> np.ndarray:
    n_body = -1
    for i, _body in enumerate(_bodies):
        if _body.name == body.name: n_body = i
    
    if n_body == -1: raise Exception("Couldn't find the body in bodies")

    @jit
    def fn(_body, f, bodies, t, delta, limit, n):
        if bodies[n_body]["n"] == _body["n"] and n < time:
            r_prev = _body["r"][n - 1]
            v_prev = _body["v"][n - 1]

            difference = destination - r_prev
            distance = ax.vector_magnitude_jit(difference)
            deceleration_dist = (speed_max ** 2) / (2 * force_max)
            distance_from_deceleration = distance - deceleration_dist
            unit_vector = ax.unit_vector_jit(difference)
            
            quaternion = axana.quaternion_between(unit_vector, v_prev)
            quaternion = axana.clip_quaternion_degrees(quaternion, turn_rate)
            target_unit_vector = axana.apply_quaternion(v_prev, quaternion)

            if distance_from_deceleration <= 0:
                force = target_unit_vector * -force_max
                force = np.clip(f, -force_max, force_max)
            else:
                force = target_unit_vector * force_max
            
            force = np.clip(f, -force_max, force_max)

            f = force + f

        return f
    return fn



import axinite.tools as axtools
import axinite as ax
from vpython import *
from itertools import cycle
import os, signal, time, threading
  
def to_vec(v): 
    return vec(v[0], v[1], v[2])
def to_vec_cartesian(v):
    return vec(v.x.value, v.y.value, v.z.value)

def vpython_frontend(args: axtools.AxiniteArgs, mode: str, **kwargs):
    """
    Initializes the VPython frontend for visualizing the simulation.

    Args:
        args (axtools.AxiniteArgs): The arguments for the simulation.
        mode (str): The mode of the frontend. Can be 'live', 'run', or 'show'.
        **kwargs: Additional keyword arguments for the frontend.

    Raises:
        Exception: If the mode is not 'live', 'run', or 'show'.

    Returns:
        function: A function to update the visualization in 'live' or 'run' mode.
        function: A function to stop the visualization in 'live' or 'run' mode.
    """
    if mode == "live" or mode == "run":
        return vpython_live(args, **kwargs)
    elif mode == "show":
        return vpython_static(args, **kwargs)
    
def vpython_live(args: axtools.AxiniteArgs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    scene = canvas(title=args.name)
    scene.select()

    colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

    global pause
    pause = False
    def pause_handler():
        global pause
        pause = not pause

    global _rate
    _rate = args.rate
    def rate_hander(evt):
        global _rate
        _rate = evt.value
    button(text='Pause', bind=pause_handler, pos=scene.caption_anchor)
    slider(bind=rate_hander, min=1, max=1000, value=+_rate, step=1, right=15, length=200, pos=scene.caption_anchor)

    spheres = {}
    labels = {}
    lights = {}

    for body in args.bodies:
        body_color = axtools.string_to_color(body.color, "vpython") if body.color != "" else next(colors)
        body_retain = body.retain if body.retain != None else args.retain
        spheres[body.name] = sphere(pos=to_vec(body.r(0)), radius=body.radius * args.radius_multiplier, color=body_color, make_trail=True, retain=body_retain, interval=10)
        labels[body.name] = label(pos=spheres[body.name].pos, text=body.name, xoffset=15, yoffset=15, space=30, height=10, border=4, font='sans')
        if body.light == True: 
            lights[body.name] = local_light(pos=to_vec(body.r(0)), color=body_color)
            attach_light(spheres[body.name], lights[body.name])

    def fn(bodies, t, **kwargs):
        bodies = ax.create_outer_bodies(bodies, kwargs["limit"], kwargs["delta"])
        global _rate, pause
        rate(_rate)
        for body in bodies:
            spheres[body.name].pos = to_vec(body.r(kwargs["n"]))
            labels[body.name].pos = spheres[body.name].pos
            try: lights[body.name].pos = spheres[body.name].pos
            except: pass
        if kwargs['n'] % 200 == 1:print(f"t = {t}, n = {kwargs['n']}", end='\r')
        if pause: 
            while pause: rate(10)

    return fn, lambda: os.kill(os.getpid(), signal.SIGINT)

def vpython_static(args: axtools.AxiniteArgs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200
       
    scene = canvas(title=args.name)
    scene.select()

    colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

    def fn1(body: axtools.Body):
        body_color = axtools.string_to_color(body.color, "vpython") if body.color != "" else next(colors)
        
        label(pos=to_vec(body.r(0)), text=body.name, xoffset=15, yoffset=15, space=30, height=10, border=4, font='sans', color=body_color)
        curve(pos=[to_vec(r) for r in body.rs], color=body_color)
        sphere(pos=to_vec(body.r(0)), radius=body.radius * args.radius_multiplier * body.radius_multiplier, color=body_color, opacity=0.2, make_trail=False)

    def exit_func():
        os.kill(os.getpid(), signal.SIGINT)

    def fn2():
        threading.Thread(target=lambda: time.sleep(1), daemon=True).start()
        signal.signal(signal.SIGINT, lambda *args, **kwargs: exit_func())
        while True: 
            rate(10)
            time.sleep(0.1)

    return fn1, fn2, lambda: exit_func()


import axinite as ax
import axinite.analysis as axana
import numpy as np
from numba import jit

@jit
def _collision_detection(bodies: np.ndarray, radii: np.ndarray, delta: np.float64) -> np.ndarray:
    """
    Detect collisions between bodies within a given time step interval.

    Args:
        bodies (np.ndarray): Array of body representations.
        radii (np.ndarray): Array of radii for each body.
        delta (np.float64): Time step interval.

    Returns:
        np.ndarray: Array of detected collisions.
    """
    collisions = []
    n = 0
    n_total = bodies["r"].shape[0]
    limit = n_total * delta
    body_dtype = ax.body_dtype(limit, delta)
    collision_dtype = np.dtype([
        ("t", np.float64),
        ("n", np.int64),
        ("r", np.float64, 3),
        ("a", body_dtype),
        ("b", body_dtype)
    ])

    for i in range(n_total):
        for j in range(i + 1, n_total):
            dist = np.linalg.norm(bodies["r"][i] - bodies["r"][j])
            if dist < (radii[i] + radii[j]):
                collision = np.zeros(1, dtype=collision_dtype)
                collision["t"] = delta
                collision["n"] = n
                collision["r"] = (bodies["r"][i] + bodies["r"][j]) / 2
                collision["a"] = bodies[i]
                collision["b"] = bodies[j]
                collisions.append(collision)
                n += 1
            else:
                for t in np.arange(0, delta, delta / 10):
                    r_i = axana.approximate_jit(bodies[i], t, delta, axana.linear_interpolation)
                    r_j = axana.approximate_jit(bodies[j], t, delta, axana.linear_interpolation)
                    dist = np.linalg.norm(r_i - r_j)
                    if dist < (radii[i] + radii[j]):
                        collision = np.zeros(1, dtype=collision_dtype)
                        collision["t"] = t
                        collision["n"] = n
                        collision["r"] = (r_i + r_j) / 2
                        collision["a"] = bodies[i]
                        collision["b"] = bodies[j]
                        collisions.append(collision)
                        n += 1
                        break

    return np.array(collisions, dtype=collision_dtype)

def collision_detection(bodies: list[ax.Body], radii: np.ndarray, delta: np.float64) -> np.ndarray:
    """
    Detect collisions between bodies within a given time step interval.

    Args:
        bodies (list[ax.Body]): List of body objects.
        radii (np.ndarray): Array of radii for each body.
        delta (np.float64): Time step interval.

    Returns:
        np.ndarray: Array of detected collisions.
    """
    return _collision_detection(ax.get_inner_bodies(bodies), radii, delta)



import axinite as ax
from typing import Literal
import vpython as vp
import numpy as np
import axinite.tools as axtools
from numba import njit

def data_to_body(data: dict[str, any], limit, delta) -> axtools.Body:
    """Converts a dict to a Body object.

    Args:
        data (dict[str, any]): The dict to convert.

    Returns:
        axtools.Body: The Body object.
    """

    name = data["name"]
    mass = ax.interpret_mass(data["mass"])

    position = np.array([
        ax.interpret_distance(data["r"][0][0]),
        ax.interpret_distance(data["r"][0][1]),
        ax.interpret_distance(data["r"][0][2])  
    ])
    velocity = np.array([
        ax.interpret_distance(data["v"][0][0]),
        ax.interpret_distance(data["v"][0][1]),
        ax.interpret_distance(data["v"][0][2])
    ])
    body = axtools.Body(name, mass, limit, delta, position, velocity)
    body.radius = ax.interpret_distance(data["radius"])

    if "color" in data:
        body.color = data["color"]
    if "light" in data:
        body.light = data["light"]
    if "retain" in data:
        body.retain = data["retain"]

    for i in range(len(data["r"])):
        body._inner["r"][i] = np.array([
            ax.interpret_distance(data["r"][i][0]),
            ax.interpret_distance(data["r"][i][1]),
            ax.interpret_distance(data["r"][i][2])
        ])
    for i in range(len(data["v"])):
        body._inner["v"][i] = np.array([
            ax.interpret_distance(data["v"][i][0]),
            ax.interpret_distance(data["v"][i][1]),
            ax.interpret_distance(data["v"][i][2])
        ])

    return body

def string_to_color(color_name: str, frontend: Literal['vpython', 'mpl', 'plotly']) -> vp.color | str:
    """Converts a string to a color object for a given frontend.

    Args:
        color_name (str): The name of the color.
        frontend (str): The frontend to convert for.

    Returns:
        vp.color | str: The converted color.
    """
    if frontend == "vpython":
        color_map = {
            'red': vp.color.red,
            'blue': vp.color.blue,
            'green': vp.color.green,
            'orange': vp.color.orange,
            'purple': vp.color.purple,
            'yellow': vp.color.yellow,
            'white': vp.color.white,
            'gray': vp.color.gray(0.5)
        }
        return color_map.get(color_name, vp.color.white)
    elif frontend == "matplotlib":
        color_map = {
            'red': 'r',
            'blue': 'b',
            'green': 'g',
            'orange': 'orange',
            'purple': 'purple',
            'yellow': 'yellow',
            'white': 'white',
            'gray': 'gray'
        }
        return color_map.get(color_name, 'white')

def create_sphere(pos: np.ndarray, radius: np.float64, n=20) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates the vertices of a sphere.

    Args:
        pos (np.ndarray): The position of the sphere.
        radius (np.float64): The radius of the sphere.
        n (int, optional): Number of segments used to generate verticies. Defaults to 20.

    Returns:
        tuple[np.ndarray, np.ndarray, np.ndarray]: The x, y, and z coordinates of the sphere.
    """
    u1 = np.linspace(0, 2 * np.pi, n)
    v1 = u1.copy()
    uu, vv = np.meshgrid(u1, v1)

    xx = pos[0] + radius * np.cos(uu) * np.sin(vv)
    yy = pos[0] + radius * np.sin(uu) * np.sin(vv)
    zz = pos[0] + radius * np.cos(vv)

    return xx, yy, zz

def sphere_has(point: np.ndarray, sphere_pos: np.ndarray, radius: np.float64) -> bool:
    """Checks if a point is inside a sphere.

    Args:
        point (np.ndarray): The point to check.
        sphere_pos (np.ndarray): The position of the sphere.
        radius (np.float64): The radius of the sphere.

    Returns:
        bool: True if the point is inside the sphere, False otherwise.
    """
    distance = np.linalg.norm(point - sphere_pos)
    return distance <= radius

def max_axis_length(*bodies: axtools.Body, radius_multiplier: int = 1) -> np.float64:
    """Finds the maximum axis length of a set of bodies.

    Args:
        radius_multiplier (int, optional): The radius multiplier to apply. Defaults to 1.

    Returns:
        np.float64: The longest axis length found.
    """

    max_length = 0
    for body in bodies:
        x_length = max([v[0] for v in body._inner["r"]]) + body.radius * radius_multiplier
        y_length = max([v[1] for v in body._inner["r"]]) + body.radius * radius_multiplier
        z_length = max([v[2] for v in body._inner["r"]]) + body.radius * radius_multiplier
        
        max_length = max(max_length, x_length, y_length, z_length)
    
    return max_length

def min_axis_length(*bodies: axtools.Body, radius_multiplier: int = 1) -> np.float64:
    """Finds the minimum axis length of a set of bodies.

    Args:
        radius_multiplier (int, optional): The radius multiplier to apply. Defaults to 1.

    Returns:
        np.float64: The lowest axis length found.
    """
    
    min_length = 0
    for body in bodies:
        x_length = min([v[0] for v in body._inner["r"]]) - body.radius * radius_multiplier
        y_length = min([v[1] for v in body._inner["r"]]) - body.radius * radius_multiplier
        z_length = min([v[2] for v in body._inner["r"]]) - body.radius * radius_multiplier
        
        min_length = min(min_length, x_length, y_length, z_length)
    
    return min_length

def from_body(body: ax.Body) -> axtools.Body:
    """Converts an ax.Body object to an axtools.Body object.

    Args:
        body (ax.Body): The Body object to convert.

    Returns:
        axtools.Body: The converted Body object.
    """

    _body = axtools.Body(body.name, body.mass, body.limit, body.delta, body._inner["r"][0], body._inner["v"][0])

    for i, r in enumerate(body._inner["r"]):
        _body._inner["r"][i] = r
    for i, v in enumerate(body._inner["v"]):
        _body._inner["v"][i] = v

    return _body


import axinite as ax
import axinite.analysis as axana
import numpy as np
from numba import jit

@jit
def _can_see(a, b, bodies, radii, ai, bi):
    unit_vector = ax.unit_vector_jit(b["r"] - a["r"])
    distance_ab = np.linalg.norm(b["r"] - a["r"])

    for i, body in enumerate(bodies):
        if i == ai or i == bi:
            continue

        vector_ac = body["r"] - a["r"]

        projection_length = np.dot(vector_ac, unit_vector)
        closest_point = a["r"] + projection_length * unit_vector

        if 0 < projection_length < distance_ab:
            distance_to_line = np.linalg.norm(body["r"] - closest_point)
            if distance_to_line < radii[i]:
                return False

    return True

def can_see(a, b, bodies, radii):
    ai = next(i for i, body in enumerate(bodies) if np.array_equal(body["r"], a["r"]))
    bi = next(i for i, body in enumerate(bodies) if np.array_equal(body["r"], b["r"]))
    return _can_see(a, b, bodies, radii, ai, bi)


import numpy as np
import axinite as ax

def magnetic_force_modifier(charges, magnetic_field):
    """
    Returns a modifier function to calculate the magnetic force between bodies in a magnetic field.

    Args:
        charges (np.ndarray): An array mapping body indexes to their charges.
        magnetic_field (np.ndarray): The magnetic field vector.

    Returns:
        function: A modifier function to calculate the magnetic force.
    """
    def modifier(body, f, bodies, t, delta, limit, n):
        charge = charges[ax.locate_body(body, bodies)]

        for i, other in enumerate(bodies):
            other_charge = charges[i]
            if other_charge == 0:
                continue

            distance_vector = other["r"][n-1] - body["r"][n-1]
            distance = np.linalg.norm(distance_vector)

            force = charge * other_charge * np.cross(magnetic_field, distance_vector) / distance**2
            f += force
        
        return f
    return modifier
